/*
** Modular Logfile Analyzer
** Copyright 2000 Jan Kneschke <jan@kneschke.de>
**
** Homepage: http://www.kneschke.de/projekte/modlogan
**

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version, and provided that the above
    copyright and permission notice is included with all distributed
    copies of this or derived software.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA

**
** $Id: plugin_config.c,v 1.7 2001/12/30 23:05:44 ostborn Exp $
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#include <ctype.h>
#include <errno.h>

#include "mlocale.h"
#include "mplugins.h"
#include "mrecord.h"
#include "mdatatypes.h"
#include "misc.h"

#include "plugin_config.h"

/* this code is taken from the squid plugin */

/* init the plugin */

int mplugins_input_qmail_dlinit(mconfig *ext_conf) {
	config_input *conf = NULL;
	const char *errptr;
	int erroffset = 0;

/* get the neccesary space */
	conf = malloc(sizeof(config_input));
	memset(conf, 0, sizeof(config_input));

/* set some defaults */	
	conf->inputfilename = NULL;
	
	conf->inputfile = stdin;
	
	conf->buf_len = 256;
	conf->buf_inc = 128;
	conf->buffer = malloc(conf->buf_len * sizeof(char));
	
	if ((conf->match_qmail_syslog_pre = pcre_compile(
		"^.{15} (.+) qmail: ",
		0, &errptr, &erroffset, NULL)) == NULL) {
		
		fprintf(stderr, "%s.%d: rexexp compilation error at %s\n", __FILE__, __LINE__, errptr);
		return -1;
	} 

/* compile the match */	
	if ((conf->match_qmail_new = pcre_compile(
		"([0-9]+)\\.([0-9]+) new msg ([0-9]+)$",
		0, &errptr, &erroffset, NULL)) == NULL) {
		
		fprintf(stderr, "%s.%d: rexexp compilation error at %s\n", __FILE__, __LINE__, errptr);
		return -1;
	}
	if ((conf->match_qmail_end = pcre_compile(
		"([0-9]+)\\.([0-9]+) end msg ([0-9]+)$",
		0, &errptr, &erroffset, NULL)) == NULL) {
		
		fprintf(stderr, "%s.%d: rexexp compilation error at %s\n", __FILE__, __LINE__, errptr);
		return -1;
	} 
	if ((conf->match_qmail_delivery = pcre_compile(
		"([0-9]+)\\.([0-9]+) delivery ([0-9]+): (success|failure|deferral): (.+)$",
		0, &errptr, &erroffset, NULL)) == NULL) {
		
		fprintf(stderr, "%s.%d: rexexp compilation error at %s\n", __FILE__, __LINE__, errptr);
		return -1;
	} 
	if ((conf->match_qmail_status = pcre_compile(
		"([0-9]+)\\.([0-9]+) status: (?:local ([0-9]+)/([0-9]+) remote ([0-9]+)/([0-9]+)(?: exitasap|)|exiting)$",
		0, &errptr, &erroffset, NULL)) == NULL) {
		
		fprintf(stderr, "%s.%d: rexexp compilation error at %s\n", __FILE__, __LINE__, errptr);
		return -1;
	} 
	if ((conf->match_qmail_starting = pcre_compile(
		"([0-9]+)\\.([0-9]+) starting delivery ([0-9]+): msg ([0-9]+) to (remote|local) (.+)$",
		0, &errptr, &erroffset, NULL)) == NULL) {
		
		fprintf(stderr, "%s.%d: rexexp compilation error at %s\n", __FILE__, __LINE__, errptr);
		return -1;
	} 
	if ((conf->match_qmail_info = pcre_compile(
		"([0-9]+)\\.([0-9]+) info msg ([0-9]+): bytes ([0-9]+) from <(.*)> qp ([0-9]+) uid ([0-9]+)$",
		0, &errptr, &erroffset, NULL)) == NULL) {
		
		fprintf(stderr, "%s.%d: rexexp compilation error at %s\n", __FILE__, __LINE__, errptr);
		return -1;
	} 
	if ((conf->match_qmail_bounce = pcre_compile(
		"([0-9]+)\\.([0-9]+) bounce msg ([0-9]+) qp ([0-9]+)",
		0, &errptr, &erroffset, NULL)) == NULL) {
		
		fprintf(stderr, "%s.%d: rexexp compilation error at %s\n", __FILE__, __LINE__, errptr);
		return -1;
	}
	if ((conf->match_qmail_triple_bounce = pcre_compile(
		"([0-9]+)\\.([0-9]+) triple bounce: discarding bounce/([0-9]+)",
		0, &errptr, &erroffset, NULL)) == NULL) {
		
		fprintf(stderr, "%s.%d: rexexp compilation error at %s\n", __FILE__, __LINE__, errptr);
		return -1;
	}
	if ((conf->match_qmail_status_smtp = pcre_compile(
		"_([0-9]{3})_",
		0, &errptr, &erroffset, NULL)) == NULL) {
		
		fprintf(stderr, "%s.%d: rexexp compilation error at %s\n", __FILE__, __LINE__, errptr);
		return -1;
	}
	if ((conf->match_qmail_status_dsn = pcre_compile(
		"_([245])\\.([0-9])\\.([0-9])_",
		0, &errptr, &erroffset, NULL)) == NULL) {
		
		fprintf(stderr, "%s.%d: rexexp compilation error at %s\n", __FILE__, __LINE__, errptr);
		return -1;
	} 
	
/* connect the plugin config to the master config */	
	ext_conf->plugin_conf = conf;
	
	return 0;
}

/* destructor */

int mplugins_input_qmail_dlclose(mconfig *ext_conf) {
	config_input *conf = ext_conf->plugin_conf;

/* clean up everything you have opened, reserved */
	if (conf->inputfilename && strcmp(conf->inputfilename, "-")) {
		fclose(conf->inputfile);
	}
	
	pcre_free(conf->match_qmail_syslog_pre);
	pcre_free(conf->match_qmail_new);
	pcre_free(conf->match_qmail_end);
	pcre_free(conf->match_qmail_delivery);
	pcre_free(conf->match_qmail_status);
	pcre_free(conf->match_qmail_starting);
	pcre_free(conf->match_qmail_info);
	pcre_free(conf->match_qmail_bounce);
	pcre_free(conf->match_qmail_status_smtp);
	pcre_free(conf->match_qmail_status_dsn);
	
	free(conf->buffer);
	
	free(ext_conf->plugin_conf);
	ext_conf->plugin_conf = NULL;
	
	return 0;
}

int mplugins_input_qmail_parse_config(mconfig *ext_conf, const char *filename, const char *section) {
	config_input *conf = ext_conf->plugin_conf;
	
	const mconfig_values config_values[] = {
		{"inputfile",	M_CONFIG_TYPE_STRING,	M_CONFIG_VALUE_OVERWRITE, &(conf->inputfilename)},
		
/* terminate the array of options */		
		{NULL, M_CONFIG_TYPE_INT, 0, NULL}
	};
	
	return mconfig_parse_section(ext_conf, filename, section, config_values);
}


int mplugins_input_qmail_set_defaults(mconfig *ext_conf) {
	config_input *conf = ext_conf->plugin_conf;

	if (conf->inputfilename && strcmp(conf->inputfilename, "-")) {
		if (!(conf->inputfile = fopen(conf->inputfilename, "r"))) {
			fprintf(stderr, "%s %s: %s\n", _("Can't open inputfile "), conf->inputfilename, strerror(errno));
			return -1;
		}
	}
	
	return 0;
}


int mplugins_init(mplugin *func) {
	func->dlinit = mplugins_input_qmail_dlinit;
	func->dlclose = mplugins_input_qmail_dlclose;
	func->parse_config = mplugins_input_qmail_parse_config;
	func->set_defaults = mplugins_input_qmail_set_defaults;
	func->get_next_record = mplugins_input_qmail_get_next_record;
	func->insert_record = NULL;
	func->gen_report = NULL;
        func->gen_history = NULL;
	 
	return 0;
}
